@@ -1,8 +1,6 @@
 """Django Application configuration."""
-from __future__ import absolute_import, unicode_literals
-
 from django.apps import AppConfig
-from django.utils.translation import ugettext_lazy as _
+from django.utils.translation import gettext_lazy as _
 
 __all__ = ['BeatConfig']
 
