@@ -1,33 +1,31 @@
 """Beat Scheduler Implementation."""
-from __future__ import absolute_import, unicode_literals
-
+import datetime
 import logging
+import math
 
 from multiprocessing.util import Finalize
 
 from celery import current_app
 from celery import schedules
 from celery.beat import Scheduler, ScheduleEntry
-from celery.five import values, items
-from celery.utils.encoding import safe_str, safe_repr
+
 from celery.utils.log import get_logger
+from celery.utils.time import maybe_make_aware
+from kombu.utils.encoding import safe_str, safe_repr
 from kombu.utils.json import dumps, loads
 
+from django.conf import settings
 from django.db import transaction, close_old_connections
-from django.db.utils import DatabaseError
+from django.db.utils import DatabaseError, InterfaceError
 from django.core.exceptions import ObjectDoesNotExist
 
 from .models import (
     PeriodicTask, PeriodicTasks,
     CrontabSchedule, IntervalSchedule,
-    SolarSchedule,
+    SolarSchedule, ClockedSchedule
 )
-from .utils import make_aware
-
-try:
-    from celery.utils.time import is_naive
-except ImportError:  # pragma: no cover
-    from celery.utils.timeutils import is_naive  # noqa
+from .clockedschedule import clocked
+from .utils import NEVER_CHECK_TIMEOUT
 
 # This scheduler must wake up more frequently than the
 # regular of 5 minutes because it needs to take external
@@ -39,7 +37,7 @@ Cannot add entry %r to database schedule: %r. Contents: %r
 """
 
 logger = get_logger(__name__)
-debug, info = logger.debug, logger.info
+debug, info, warning = logger.debug, logger.info, logger.warning
 
 
 class ModelEntry(ScheduleEntry):
@@ -49,6 +47,7 @@ class ModelEntry(ScheduleEntry):
         (schedules.crontab, CrontabSchedule, 'crontab'),
         (schedules.schedule, IntervalSchedule, 'interval'),
         (schedules.solar, SolarSchedule, 'solar'),
+        (clocked, ClockedSchedule, 'clocked')
     )
     save_fields = ['last_run_at', 'total_run_count', 'no_changes']
 
@@ -75,18 +74,25 @@ class ModelEntry(ScheduleEntry):
             )
             self._disable(model)
 
-        self.options = {
-            'queue': model.queue,
-            'exchange': model.exchange,
-            'routing_key': model.routing_key,
-            'expires': model.expires,
-        }
+        self.options = {}
+        for option in ['queue', 'exchange', 'routing_key', 'priority']:
+            value = getattr(model, option)
+            if value is None:
+                continue
+            self.options[option] = value
+
+        if getattr(model, 'expires_', None):
+            self.options['expires'] = getattr(model, 'expires_')
+
+        self.options['headers'] = loads(model.headers or '{}')
+
         self.total_run_count = model.total_run_count
         self.model = model
 
         if not model.last_run_at:
             model.last_run_at = self._default_now()
-        self.last_run_at = make_aware(model.last_run_at)
+
+        self.last_run_at = model.last_run_at
 
     def _disable(self, model):
         model.no_changes = True
@@ -95,18 +101,54 @@ class ModelEntry(ScheduleEntry):
 
     def is_due(self):
         if not self.model.enabled:
-            return False, 5.0   # 5 second delay for re-enable.
-        return self.schedule.is_due(self.last_run_at)
+            # 5 second delay for re-enable.
+            return schedules.schedstate(False, 5.0)
+
+        # START DATE: only run after the `start_time`, if one exists.
+        if self.model.start_time is not None:
+            now = self._default_now()
+            if getattr(settings, 'DJANGO_CELERY_BEAT_TZ_AWARE', True):
+                now = maybe_make_aware(self._default_now())
+
+            if now < self.model.start_time:
+                # The datetime is before the start date - don't run.
+                # send a delay to retry on start_time
+                delay = math.ceil(
+                    (self.model.start_time - now).total_seconds()
+                )
+                return schedules.schedstate(False, delay)
+
+        # ONE OFF TASK: Disable one off tasks after they've ran once
+        if self.model.one_off and self.model.enabled \
+                and self.model.total_run_count > 0:
+            self.model.enabled = False
+            self.model.total_run_count = 0  # Reset
+            self.model.no_changes = False  # Mark the model entry as changed
+            self.model.save()
+            # Don't recheck
+            return schedules.schedstate(False, NEVER_CHECK_TIMEOUT)
+
+        # CAUTION: make_aware assumes settings.TIME_ZONE for naive datetimes,
+        # while maybe_make_aware assumes utc for naive datetimes
+        tz = self.app.timezone
+        last_run_at_in_tz = maybe_make_aware(self.last_run_at).astimezone(tz)
+        return self.schedule.is_due(last_run_at_in_tz)
 
     def _default_now(self):
-        now = self.app.now()
         # The PyTZ datetime must be localised for the Django-Celery-Beat
         # scheduler to work. Keep in mind that timezone arithmatic
         # with a localized timezone may be inaccurate.
-        return now.tzinfo.localize(now.replace(tzinfo=None))
+        if getattr(settings, 'DJANGO_CELERY_BEAT_TZ_AWARE', True):
+            now = self.app.now()
+            now = now.tzinfo.localize(now.replace(tzinfo=None))
+        else:
+            # this ends up getting passed to maybe_make_aware, which expects
+            # all naive datetime objects to be in utc time.
+            now = datetime.datetime.utcnow()
+        return now
 
     def __next__(self):
-        self.model.last_run_at = self.app.now()
+        self.model.last_run_at = self._default_now()
         self.model.total_run_count += 1
         self.model.no_changes = True
         return self.__class__(self.model)
@@ -118,6 +160,7 @@ class ModelEntry(ScheduleEntry):
         obj = type(self.model)._default_manager.get(pk=self.model.pk)
         for field in self.save_fields:
             setattr(obj, field, getattr(self.model, field))
+
         obj.save()
 
     @classmethod
@@ -152,11 +195,15 @@ class ModelEntry(ScheduleEntry):
 
     @classmethod
     def _unpack_options(cls, queue=None, exchange=None, routing_key=None,
+                        priority=None, headers=None, expire_seconds=None,
                         **kwargs):
         return {
             'queue': queue,
             'exchange': exchange,
             'routing_key': routing_key,
+            'priority': priority,
+            'headers': dumps(headers or {}),
+            'expire_seconds': expire_seconds,
         }
 
     def __repr__(self):
@@ -175,7 +222,8 @@ class DatabaseScheduler(Scheduler):
 
     _schedule = None
     _last_timestamp = None
-    _initial_read = False
+    _initial_read = True
+    _heap_invalidated = False
 
     def __init__(self, *args, **kwargs):
         """Initialize the database scheduler."""
@@ -183,9 +231,9 @@ class DatabaseScheduler(Scheduler):
         Scheduler.__init__(self, *args, **kwargs)
         self._finalize = Finalize(self, self.sync, exitpriority=5)
         self.max_interval = (
-            kwargs.get('max_interval') or
-            self.app.conf.beat_max_loop_interval or
-            DEFAULT_MAX_INTERVAL)
+            kwargs.get('max_interval')
+            or self.app.conf.beat_max_loop_interval
+            or DEFAULT_MAX_INTERVAL)
 
     def setup_schedule(self):
         self.install_default_entries(self.schedule)
@@ -203,6 +251,8 @@ class DatabaseScheduler(Scheduler):
 
     def schedule_changed(self):
         try:
+            close_old_connections()
+
             # If MySQL is running with transaction isolation level
             # REPEATABLE-READ (default), then we won't see changes done by
             # other transactions until the current transaction is
@@ -216,6 +266,13 @@ class DatabaseScheduler(Scheduler):
         except DatabaseError as exc:
             logger.exception('Database gave error: %r', exc)
             return False
+        except InterfaceError:
+            warning(
+                'DatabaseScheduler: InterfaceError in schedule_changed(), '
+                'waiting to retry in next call...'
+            )
+            return False
+
         try:
             if ts and ts > (last if last else ts):
                 return True
@@ -231,26 +288,34 @@ class DatabaseScheduler(Scheduler):
         return new_entry
 
     def sync(self):
-        info('Writing entries...')
+        if logger.isEnabledFor(logging.DEBUG):
+            debug('Writing entries...')
         _tried = set()
+        _failed = set()
         try:
             close_old_connections()
-            with transaction.atomic():
-                while self._dirty:
-                    try:
-                        name = self._dirty.pop()
-                        _tried.add(name)
-                        self.schedule[name].save()
-                    except (KeyError, ObjectDoesNotExist):
-                        pass
+
+            while self._dirty:
+                name = self._dirty.pop()
+                try:
+                    self.schedule[name].save()
+                    _tried.add(name)
+                except (KeyError, ObjectDoesNotExist):
+                    _failed.add(name)
         except DatabaseError as exc:
-            # retry later
-            self._dirty |= _tried
             logger.exception('Database error while sync: %r', exc)
+        except InterfaceError:
+            warning(
+                'DatabaseScheduler: InterfaceError in sync(), '
+                'waiting to retry in next call...'
+            )
+        finally:
+            # retry later, only for the failed ones
+            self._dirty |= _failed
 
     def update_from_dict(self, mapping):
         s = {}
-        for name, entry_fields in items(mapping):
+        for name, entry_fields in mapping.items():
             try:
                 entry = self.Entry.from_entry(name,
                                               app=self.app,
@@ -259,7 +324,7 @@ class DatabaseScheduler(Scheduler):
                     s[name] = entry
 
             except Exception as exc:
-                logger.error(ADD_ENTRY_ERROR, name, exc, entry_fields)
+                logger.exception(ADD_ENTRY_ERROR, name, exc, entry_fields)
         self.schedule.update(s)
 
     def install_default_entries(self, data):
@@ -269,18 +334,24 @@ class DatabaseScheduler(Scheduler):
                 'celery.backend_cleanup', {
                     'task': 'celery.backend_cleanup',
                     'schedule': schedules.crontab('0', '4', '*'),
-                    'options': {'expires': 12 * 3600},
+                    'options': {'expire_seconds': 12 * 3600},
                 },
             )
         self.update_from_dict(entries)
 
+    def schedules_equal(self, *args, **kwargs):
+        if self._heap_invalidated:
+            self._heap_invalidated = False
+            return False
+        return super(DatabaseScheduler, self).schedules_equal(*args, **kwargs)
+
     @property
     def schedule(self):
-        update = False
-        if not self._initial_read:
+        initial = update = False
+        if self._initial_read:
             debug('DatabaseScheduler: initial read')
-            update = True
-            self._initial_read = True
+            initial = update = True
+            self._initial_read = False
         elif self.schedule_changed():
             info('DatabaseScheduler: Schedule changed.')
             update = True
@@ -289,9 +360,11 @@ class DatabaseScheduler(Scheduler):
             self.sync()
             self._schedule = self.all_as_schedule()
             # the schedule changed, invalidate the heap in Scheduler.tick
-            self._heap = None
+            if not initial:
+                self._heap = []
+                self._heap_invalidated = True
             if logger.isEnabledFor(logging.DEBUG):
                 debug('Current schedule:\n%s', '\n'.join(
-                    repr(entry) for entry in values(self._schedule)),
+                    repr(entry) for entry in self._schedule.values()),
                 )
         return self._schedule
