@@ -1,12 +1,12 @@
 """Utilities."""
 # -- XXX This module must not use translation as that causes
 # -- a recursive loader import!
-from __future__ import absolute_import, unicode_literals
-
 from django.conf import settings
 from django.utils import timezone
 
 is_aware = timezone.is_aware
+# celery schedstate return None will make it not work
+NEVER_CHECK_TIMEOUT = 100000000
 
 # see Issue #222
 now_localtime = getattr(timezone, 'template_localtime', timezone.localtime)
@@ -21,6 +21,10 @@ def make_aware(value):
         # then convert to the Django configured timezone.
         default_tz = timezone.get_default_timezone()
         value = timezone.localtime(value, default_tz)
+    else:
+        # naive datetimes are assumed to be in local timezone.
+        if timezone.is_naive(value):
+            value = timezone.make_aware(value, timezone.get_default_timezone())
     return value
 
 
@@ -39,6 +43,6 @@ def is_database_scheduler(scheduler):
     from kombu.utils import symbol_by_name
     from .schedulers import DatabaseScheduler
     return (
-        scheduler == 'django' or
-        issubclass(symbol_by_name(scheduler), DatabaseScheduler)
+        scheduler == 'django'
+        or issubclass(symbol_by_name(scheduler), DatabaseScheduler)
     )
